# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
import subprocess
import sys
from argparse import ArgumentParser, FileType, Namespace
from datetime import date
from pathlib import Path
from typing import Dict, Iterable, List, Union

import tomlkit
from tomlkit.toml_document import TOMLDocument

from pontos.helper import shell_cmd_runner
from pontos.release.helper import get_project_name
from pontos.terminal import Terminal
from pontos.terminal.null import NullTerminal
from pontos.terminal.rich import RichTerminal

ADDRESS = "https://github.com/"


class ChangelogBuilder:
    """Creates Changelog files from conventional commits
    using the git log, from the latest tag"""

    def __init__(
        self,
        terminal: Terminal,
        args: Namespace,
    ):
        self._terminal = terminal
        self.config: TOMLDocument = tomlkit.parse(
            args.config.read_text(encoding="utf-8")
        )
        self.project: str = (
            args.project
            if args.project is not None
            else get_project_name(shell_cmd_runner)
        )
        self.space: str = args.space
        changelog_dir: Path = Path.cwd() / self.config.get("changelog_dir")
        changelog_dir.mkdir(parents=True, exist_ok=True)
        self.output_file: Path = changelog_dir / args.output
        self.current_version: str = args.current_version
        self.next_version: str = args.next_version

    def create_changelog_file(self) -> Union[Path, None]:
        commit_list = self._get_git_log()
        commit_dict = self._sort_commits(commit_list)
        return self._build_changelog_file(commit_dict=commit_dict)

    def _get_git_log(self) -> Union[List[str], None]:
        """Getting the git log for the current branch.

        If a last tag is found, the `git log` is searched up to the
          last tag, found by `git tag | sort -V | tail -1`.
        Else if no tag is found, the `git log` is searched full, assuming
          this is a initial release.

        Returns:
            A list of `git log` entries or None
        """
        # https://stackoverflow.com/a/12083016/6725620
        # uses only latest tag for this branch
        # catch this CalledProcessError on
        # "No names found, cannot describe anything."
        cmd: str = "git log HEAD --oneline"
        try:
            # https://gist.github.com/rponte/fdc0724dd984088606b0
            proc: subprocess.CompletedProcess = shell_cmd_runner(
                "git tag | sort -V | tail -1"
            )
        except subprocess.CalledProcessError:
            self._terminal.warning("No tag found.")

        if proc.stdout and proc.stdout != "":
            cmd: str = f'git log "{proc.stdout.strip()}..HEAD" --oneline'

        proc = shell_cmd_runner(cmd)
        if proc.stdout and proc.stdout != "":
            return proc.stdout.strip().split("\n")
        return None

    def _sort_commits(self, commits: List[str]):
        """Sort the commits by commit type and group them
        in a dict
        ```
        {
            'Added:': [
                'commit 1 [1234567](..)',
                'commit 2 [1234568](..)',
                '...',
            ],
            'Fixed:': [
                ...
            ],
        }
        ```

        Returns
            The dict containing the commit messages"""
        # get the commit types from the toml
        commit_types = self.config.get("commit_types")

        commit_link = f"{ADDRESS}{self.space}/{self.project}/commit/"

        commit_dict = {}
        if commits and len(commits) > 0:
            for commit in commits:
                commit = commit.split(" ", maxsplit=1)
                for commit_type in commit_types:
                    reg = re.compile(
                        rf'{commit_type["message"]}\s?[:|-]', flags=re.I
                    )
                    match = reg.match(commit[1])
                    if match:
                        if commit_type["group"] not in commit_dict:
                            commit_dict[commit_type["group"]] = []

                        # remove the commit tag from commit message
                        cleaned_msg = (
                            commit[1].replace(match.group(0), "").strip()
                        )
                        commit_dict[commit_type["group"]].append(
                            f"{cleaned_msg} [{commit[0]}]"
                            f"({commit_link}{commit[0]})"
                        )
                        self._terminal.info(f"{commit[0]}: {cleaned_msg}")
        if not commit_dict:
            self._terminal.warning("No conventional commits found.")
            sys.exit(1)
        return commit_dict

    def _build_changelog_file(self, commit_dict: Dict) -> Union[str, None]:
        """Building the changelog file with the passed dict.

        Arguments:
            commit_dict     dict containing sorted commits

        Returns:
            The path to the changelog file or None"""

        # changelog header
        changelog = (
            "# Changelog\n\n"
            "All notable changes to this project "
            "will be documented in this file.\n\n"
        )
        if self.next_version:
            changelog += (
                f"## [{self.next_version}] - {date.today().isoformat()}\n"
            )
        else:
            changelog += "## [Unreleased]\n\n"

        # changelog entries
        commit_types = self.config.get("commit_types")
        for commit_type in commit_types:
            if commit_type["group"] in commit_dict.keys():
                changelog += f"\n## {commit_type['group']}\n"
                for msg in commit_dict[commit_type["group"]]:
                    changelog += f"* {msg}\n"

        # comparison line (footer)
        pre = "\n[Unreleased]: "
        compare_link = f"{ADDRESS}{self.space}/{self.project}/compare/"
        if self.next_version and self.current_version:
            pre = f"\n[{self.next_version}]: "
            diff = f"{self.current_version}...{self.next_version}"
        elif self.current_version:
            diff = f"\n{self.current_version}...HEAD"
        else:
            diff = "???...HEAD"

        changelog += f"{pre}{compare_link}{diff}"

        if changelog:
            self.output_file.write_text(changelog, encoding="utf-8")
            return self.output_file
        return None


def parse_args(args: Iterable[str] = None) -> ArgumentParser:
    parser = ArgumentParser(
        description="Conventional commits utility.",
        prog="pontos-changelog",
    )

    parser.add_argument(
        "--config",
        "-C",
        default=Path("changelog.toml"),
        type=FileType("r"),
        help="Conventional commits config file (toml), including conventions.",
    )

    parser.add_argument(
        "--project",
        help="The github project",
    )

    parser.add_argument(
        "--space",
        default="greenbone",
        help="User/Team name in github",
    )

    parser.add_argument(
        "--current-version",
        default="greenbone",
        help="Current version before these changes",
    )

    parser.add_argument(
        "--next-version",
        help="The planned release version",
    )

    parser.add_argument(
        "--output",
        "-o",
        default=Path("unreleased.md"),
        type=FileType("r"),
        help="The path to the output file (.md)",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    parser.add_argument(
        "--log-file",
        dest="log_file",
        type=str,
        help="Activate logging using the given file path",
    )

    return parser.parse_args(args=args)


def main(
    args=None,
):

    parsed_args = parse_args(args)

    if parsed_args.quiet:
        term = NullTerminal()
    else:
        term = RichTerminal()

    term.bold_info("pontos-changelog")

    with term.indent():
        try:
            changelog_builder = ChangelogBuilder(
                terminal=term,
                args=parsed_args,
            )
            changelog_builder.create_changelog_file()
        except subprocess.CalledProcessError as e:
            term.error(f'Could not run command "{e.cmd}".')
            term.out(f"Error was: {e.stderr}")
            sys.exit(1)

    return sys.exit(0)



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
from pathlib import Path
from subprocess import CalledProcessError

from .helper import (
    VersionError,
    is_version_pep440_compliant,
    strip_version,
    versions_equal,
)
from .version import VersionCommand


# This class is used for Python Version command(s)
class GoVersionCommand(VersionCommand):
    def __init__(self, *, project_file_path: Path = None) -> None:
        if not project_file_path:
            project_file_path = Path.cwd() / "go.mod"

        if not project_file_path.exists():
            raise VersionError(f"{str(project_file_path)} file not found.")

        self.shell_cmd_runner = lambda x: subprocess.run(
            x,
            shell=True,
            check=True,
            errors="utf-8",  # use utf-8 encoding for error output
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        super().__init__(
            project_file_path=project_file_path,
        )

    def get_current_version(self) -> str:
        """Get the current version of this project
        In go the version is only defined within the repository
        tags, thus we need to check git, what tag is the latest"""
        try:
            proc = self.shell_cmd_runner(
                "git describe --tags `git rev-list --tags --max-count=1`"
            )
            version = strip_version(proc.stdout)
            return version if version is not None else ""
        except CalledProcessError as e:
            self._print(
                "No version tag found. Maybe this "
                "module has not been released at all."
            )
            raise e

    def verify_version(self, version: str) -> None:
        """Verify the current version of this project"""
        current_version = self.get_current_version()
        if not is_version_pep440_compliant(current_version):
            raise VersionError(
                f"The version {current_version} is not PEP 440 compliant."
            )

        if versions_equal(self.get_current_version(), version):
            self._print("OK")

    def update_version(
        self, new_version: str, *, develop: bool = False, force: bool = False
    ) -> None:
        """Update the current version of this project"""
        _ = (new_version, develop, force)
        self._print("Updating the version of a go module is not possible.")



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import ContextManager, Iterable, Iterator, Optional, Tuple, Union

import httpx

from pontos.github.api.helper import JSON_OBJECT
from pontos.helper import DownloadProgressIterable, download


class GitHubRESTReleaseMixin:
    def create_release(
        self,
        repo: str,
        tag: str,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        target_commitish: Optional[str] = None,
        draft: Optional[bool] = False,
        prerelease: Optional[bool] = False,
    ):
        """
        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            body: Content of the changelog for the release
            name: name of the release, e.g. 'pontos 1.0.0'
            target_commitish: Only needed when tag is not there yet
            draft: If the release is a draft. False by default.
            prerelease: If the release is a pre release. False by default.
        """
        data = {
            "tag_name": tag,
            "draft": draft,
            "prerelease": prerelease,
        }
        if name is not None:
            data["name"] = name
        if body is not None:
            data["body"] = body
        if target_commitish is not None:
            data["target_commitish"] = target_commitish

        api = f"/repos/{repo}/releases"
        response: httpx.Response = self._request(
            api, data=data, request=httpx.post
        )
        response.raise_for_status()

    def release_exists(self, repo: str, tag: str) -> bool:
        """
        Check wether a GitHub release exists by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Returns:
            True if the release exists
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response: httpx.Response = self._request(api)
        return response.is_success

    def release(self, repo: str, tag: str) -> JSON_OBJECT:
        """
        Get data of a GitHub release by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response: httpx.Response = self._request(api)
        response.raise_for_status()
        return response.json()

    def download_release_tarball(
        self, repo: str, tag: str, destination: Path
    ) -> ContextManager[DownloadProgressIterable]:
        """
        Download a release tarball (tar.gz) file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            destination: A path where to store the downloaded file

        Raises:
            HTTPError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.tar.gz"
        return download(api, destination)

    def download_release_zip(
        self, repo: str, tag: str, destination: Path
    ) -> ContextManager[DownloadProgressIterable]:
        """
        Download a release zip file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            destination: A path where to store the downloaded file

        Raises:
            HTTPError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.zip"
        return download(api, destination)

    def download_release_assets(
        self,
        repo: str,
        tag: str,
    ) -> Iterator[DownloadProgressIterable]:
        """
        Download release assets

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPError if the request was invalid
        """
        release_json = self.release(repo, tag)
        assets_url = release_json.get("assets_url")
        if not assets_url:
            return

        response: httpx.Response = self._request_internal(assets_url)
        response.raise_for_status()
        assets_json = response.json()
        for asset_json in assets_json:
            asset_url: str = asset_json.get("browser_download_url", "")
            name: str = asset_json.get("name", "")
            if name.endswith(".zip") or name.endswith(".tar.gz"):
                with download(
                    asset_url,
                    Path(name),
                ) as progress:
                    yield progress

    def upload_release_assets(
        self,
        repo: str,
        tag: str,
        files: Iterable[Union[Path, Tuple[Path, str]]],
    ) -> Iterator[Path]:
        # pylint: disable=line-too-long
        """
        Upload release assets one after another

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            files: An iterable of file paths or an iterable of tuples
                containing a file path and content types to upload as an asset

        Returns:
            yields each file after its upload is finished

        Raises:
            HTTPError if an upload request was invalid

        Examples:
            >>> files = (Path("foo.txt"), Path("bar.txt"),)
            >>> for uploaded_file in git.upload_release_assets("foo/bar", "1.2.3", files):
            >>>    print(f"Uploaded: {uploaded_file}")

            >>> files = [(Path("foo.txt"), "text/ascii"), (Path("bar.pdf"), "application/pdf")]
            >>> for uploaded_file in git.upload_release_assets("foo/bar", "1.2.3", files):
            >>>    print(f"Uploaded: {uploaded_file}")
        """
        github_json = self.release(repo, tag)
        asset_url = github_json["upload_url"].replace("{?name,label}", "")

        for file_path in files:
            if isinstance(file_path, Tuple):
                file_path, content_type = file_path
            else:
                content_type = "application/octet-stream"

            to_upload = file_path.read_bytes()

            response: httpx.Response = self._request_internal(
                asset_url,
                params={"name": file_path.name},
                content=to_upload,
                content_type=content_type,
                request=httpx.post,
            )
            response.raise_for_status()
            yield file_path



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from pontos.helper import DEFAULT_TIMEOUT, DownloadProgressIterable, download


class DownloadProgressIterableTestCase(unittest.TestCase):
    def test_properties(self):
        content = ["foo", "bar", "baz"]
        destination = Path("foo")
        length = 123
        url = "bar"
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=length,
            url=url,
        )

        self.assertEqual(download_progress.url, url)
        self.assertEqual(download_progress.length, length)
        self.assertEqual(download_progress.destination, destination)

    def test_progress_without_length(self):
        content = ["foo", "bar", "baz"]
        destination = MagicMock()
        writer = MagicMock()
        context_manager = MagicMock()
        context_manager.__enter__.return_value = writer
        destination.open.return_value = context_manager
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=None,
            url="foo",
        )

        self.assertEqual(download_progress.url, "foo")

        it = iter(download_progress)
        progress = next(it)
        writer.write.assert_called_with("foo")
        self.assertIsNone(progress)
        progress = next(it)
        writer.write.assert_called_with("bar")
        self.assertIsNone(progress)
        progress = next(it)
        writer.write.assert_called_with("baz")
        self.assertIsNone(progress)

        with self.assertRaises(StopIteration):
            next(it)

    def test_progress(self):
        content = ["foo", "bar", "baz"]
        destination = MagicMock()
        writer = MagicMock()
        context_manager = MagicMock()
        context_manager.__enter__.return_value = writer
        destination.open.return_value = context_manager
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=9,
            url="foo",
        )

        it = iter(download_progress)
        progress = next(it)
        writer.write.assert_called_with("foo")
        self.assertEqual(progress, 1 / 3.0)
        progress = next(it)
        writer.write.assert_called_with("bar")
        self.assertEqual(progress, 2 / 3.0)
        progress = next(it)
        writer.write.assert_called_with("baz")
        self.assertEqual(progress, 1)

        with self.assertRaises(StopIteration):
            next(it)

    def test_run(self):
        content = ["foo", "bar", "baz"]
        destination = MagicMock()
        writer = MagicMock()
        context_manager = MagicMock()
        context_manager.__enter__.return_value = writer
        destination.open.return_value = context_manager
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=9,
            url="foo",
        )

        download_progress.run()
        destination.open.assert_called_once()
        writer.write.assert_has_calls((call("foo"), call("bar"), call("baz")))


class DownloadTestCase(unittest.TestCase):
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_without_destination(
        self,
        requests_mock: MagicMock,
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        with download(
            "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz"  # pylint: disable=line-too-long
        ) as download_progress:

            requests_mock.assert_called_once_with(
                "GET",
                "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
                follow_redirects=True,
                timeout=DEFAULT_TIMEOUT,
                headers=None,
                params=None,
            )
            response_headers.get.assert_called_once_with("content-length")

            self.assertIsNone(download_progress.length)
            self.assertEqual(
                download_progress.destination, Path("v21.11.0.tar.gz")
            )

            it = iter(download_progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)

            with self.assertRaises(StopIteration):
                next(it)

            download_progress.destination.unlink()

    @patch("pontos.helper.Path")
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_with_content_length(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = "9"
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        download_file = path_mock()
        file_mock = MagicMock()
        file_mock.__enter__.return_value = file_mock
        download_file.open.return_value = file_mock

        with download(
            "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
            download_file,
        ) as download_progress:

            requests_mock.assert_called_once_with(
                "GET",
                "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
                timeout=DEFAULT_TIMEOUT,
                follow_redirects=True,
                headers=None,
                params=None,
            )
            response_headers.get.assert_called_once_with("content-length")

            self.assertEqual(download_progress.length, 9)

            it = iter(download_progress)

            progress = next(it)
            self.assertEqual(progress, 1 / 3)
            file_mock.write.assert_called_with(b"foo")

            progress = next(it)
            self.assertEqual(progress, 2 / 3)
            file_mock.write.assert_called_with(b"bar")

            progress = next(it)
            self.assertEqual(progress, 1)
            file_mock.write.assert_called_with(b"baz")

            with self.assertRaises(StopIteration):
                next(it)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines

import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

import httpx

from pontos.github.api import GitHubRESTApi, RepositoryType
from tests.github.api import default_request

here = Path(__file__).parent


class GitHubOrganizationsTestCase(unittest.TestCase):
    @patch("pontos.github.api.api.httpx.get")
    def test_organization_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.ok = True
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.organisation_exists("foo")

        args, kwargs = default_request(
            "https://api.github.com/orgs/foo",
        )
        requests_mock.assert_called_once_with(*args, **kwargs)
        self.assertTrue(exists)

    @patch("pontos.github.api.api.httpx.get")
    def test_organization_not_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.is_success = False
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.organisation_exists("foo")

        args, kwargs = default_request(
            "https://api.github.com/orgs/foo",
        )
        requests_mock.assert_called_once_with(*args, **kwargs)
        self.assertFalse(exists)

    @patch("pontos.github.api.api.httpx.get")
    def test_get_repositories(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        response1 = httpx.Response(
            status_code=200,
            json={"public_repos": 1, "total_private_repos": 1},
            request=MagicMock(),
        )
        response2 = httpx.Response(
            status_code=200,
            json=[{"foo": "bar"}, {"foo": "baz"}],
            request=MagicMock(),
        )
        requests_mock.side_effect = [response1, response2]
        ret = api.get_repositories(
            orga="foo", repository_type=RepositoryType.ALL
        )

        args, kwargs = default_request(
            "https://api.github.com/orgs/foo",
        )
        requests_mock.assert_any_call(*args, **kwargs)
        args2, kwargs2 = default_request(
            "https://api.github.com/orgs/foo/repos",
            params={"per_page": 100, "page": 1, "type": "all"},
        )
        requests_mock.assert_any_call(*args2, **kwargs2)
        self.assertEqual(ret, [{"foo": "bar"}, {"foo": "baz"}])

    @patch("pontos.github.api.api.httpx.get")
    def test_get_organization_repository_number_all(
        self, requests_mock: MagicMock
    ):
        api = GitHubRESTApi("12345")
        response = httpx.Response(
            status_code=200,
            json={"public_repos": 10, "total_private_repos": 10},
            request=MagicMock(),
        )

        requests_mock.return_value = response
        ret = api.get_organization_repository_number(
            orga="foo", repository_type=RepositoryType.ALL
        )

        args, kwargs = default_request(
            "https://api.github.com/orgs/foo",
        )
        requests_mock.assert_any_call(*args, **kwargs)

        self.assertEqual(ret, 20)

    @patch("pontos.github.api.api.httpx.get")
    def test_get_organization_repository_number_private(
        self, requests_mock: MagicMock
    ):
        api = GitHubRESTApi("12345")
        response = httpx.Response(
            status_code=200,
            json={"public_repos": 10, "total_private_repos": 10},
            request=MagicMock(),
        )

        requests_mock.return_value = response
        ret = api.get_organization_repository_number(
            orga="foo", repository_type=RepositoryType.PRIVATE
        )

        args, kwargs = default_request(
            "https://api.github.com/orgs/foo",
        )
        requests_mock.assert_any_call(*args, **kwargs)

        self.assertEqual(ret, 10)

    @patch("pontos.github.api.api.httpx.get")
    def test_get_organization_repository_number_public(
        self, requests_mock: MagicMock
    ):
        api = GitHubRESTApi("12345")
        response = httpx.Response(
            status_code=200,
            json={"public_repos": 10, "total_private_repos": 10},
            request=MagicMock(),
        )

        requests_mock.return_value = response
        ret = api.get_organization_repository_number(
            orga="foo", repository_type=RepositoryType.PUBLIC
        )

        args, kwargs = default_request(
            "https://api.github.com/orgs/foo",
        )
        requests_mock.assert_any_call(*args, **kwargs)

        self.assertEqual(ret, 10)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from typing import Dict, List, Optional, Union

import httpx

DEFAULT_GITHUB_API_URL = "https://api.github.com"
DEFAULT_TIMEOUT_CONFIG = httpx.Timeout(180.0)  # three minutes
JSON_OBJECT = Dict[str, Union[str, bool, int]]  # pylint: disable=invalid-name
JSON = Union[List[JSON_OBJECT], JSON_OBJECT]


class FileStatus(Enum):
    ADDED = "added"
    DELETED = "deleted"
    MODIFIED = "modified"
    RENAMED = "renamed"


class RepositoryType(Enum):
    ALL = "all"
    PUBLIC = "public"
    PRIVATE = "private"
    FORKS = "forks"
    SOURCES = "sources"
    MEMBER = "member"
    INTERNAL = "internal"


class WorkflowRunStatus(Enum):
    ACTION_REQUIRED = "action_required"
    CANCELLED = "cancelled"
    COMPLETED = "completed"
    FAILURE = "failure"
    IN_PROGRESS = "in_progress"
    NEUTRAL = "neutral"
    QUEUED = "queued"
    REQUESTED = "requested"
    SKIPPED = "skipped"
    STALE = "stale"
    SUCCESS = "success"
    TIMED_OUT = "timed_out"
    WAITING = "waiting"


def _get_next_url(response: httpx.Response) -> Optional[str]:
    if response and response.links:
        try:
            return response.links["next"]["url"]
        except KeyError:
            pass

    return None



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from pathlib import Path

from pontos.github.actions.event import GitHubEvent, PullRequestState

here = Path(__file__).parent


class GitHubPullRequestEventTestCase(unittest.TestCase):
    def setUp(self) -> None:
        event_path = Path(here / "test-pull-request-event.json")
        event = GitHubEvent(event_path)
        self.pull_request = event.pull_request

    def test_draft(self):
        self.assertFalse(self.pull_request.draft)

    def test_labels(self):
        self.assertEqual(len(self.pull_request.labels), 1)
        self.assertEqual(self.pull_request.labels[0].name, "enhancement")

    def test_number(self):
        self.assertEqual(self.pull_request.number, 1)

    def test_title(self):
        self.assertEqual(self.pull_request.title, "Add foo for bar")

    def test_state(self):
        self.assertEqual(self.pull_request.state, PullRequestState.OPEN)

    def test_base(self):
        base = self.pull_request.base
        self.assertEqual(base.name, "main")
        self.assertEqual(base.sha, "50c0fbe90d0c19dba08165b707e8b720d604ed5d")

    def test_head(self):
        head = self.pull_request.head
        self.assertEqual(head.name, "label-test")
        self.assertEqual(head.sha, "beeecfeee02f5a9e69c1f69dba6757df6e5c20f7")

    def test_merged(self):
        self.assertEqual(self.pull_request.merged, False)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import httpx


class GitHubRESTBranchMixin:
    def branch_exists(self, repo: str, branch: str) -> bool:
        """
        Check if a single branch in a repository exists

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch name to check
        """
        api = f"/repos/{repo}/branches/{branch}"
        response: httpx.Response = self._request(api)
        return response.is_success

    def delete_branch(self, repo: str, branch: str):
        """
        Delete a branch on GitHub

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch to be deleted

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/git/refs/{branch}"
        response: httpx.Response = self._request(api, request=httpx.delete)
        response.raise_for_status()

    def branch_protection_rules(self, repo: str, branch: str):
        """
        Get branch protection rules for a specific repository
        branch

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Get protection rules for this branch

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/branches/{branch}/protection"
        response: httpx.Response = self._request(api)
        return response.json()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines

import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

import httpx

from pontos.github.api import GitHubRESTApi
from pontos.helper import DEFAULT_TIMEOUT
from tests.github.api import default_request

here = Path(__file__).parent


class GitHubArtifactsTestCase(unittest.TestCase):
    @patch("pontos.github.api.api.httpx.get")
    def test_get_repository_artifacts(self, requests_mock: MagicMock):
        response = MagicMock()
        response.links = None
        response.json.return_value = {
            "total_count": 1,
            "artifacts": [
                {
                    "id": 11,
                    "node_id": "MDg6QXJ0aWZhY3QxMQ==",
                    "name": "Foo",
                }
            ],
        }
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        artifacts = api.get_repository_artifacts("foo/bar")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/actions/artifacts",
            params={"per_page": 100, "page": 1},
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

        self.assertEqual(len(artifacts), 1)
        self.assertEqual(artifacts[0]["name"], "Foo")

    @patch("pontos.github.api.api.httpx.get")
    def test_get_repository_artifacts_with_pagination(
        self, requests_mock: MagicMock
    ):
        response = MagicMock()
        response.links = None
        response.json.side_effect = [
            {
                "total_count": 120,
                "artifacts": [
                    {
                        "id": id,
                        "name": f"Foo-{id}",
                    }
                    for id in range(0, 100)
                ],
            },
            {
                "total_count": 120,
                "artifacts": [
                    {
                        "id": id,
                        "name": f"Foo-{id}",
                    }
                    for id in range(100, 120)
                ],
            },
        ]
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        artifacts = api.get_repository_artifacts("foo/bar")

        args1, kwargs1 = default_request(
            "https://api.github.com/repos/foo/bar/actions/artifacts",
            params={"per_page": 100, "page": 1},
        )
        args2, kwargs2 = default_request(
            "https://api.github.com/repos/foo/bar/actions/artifacts",
            params={"per_page": 100, "page": 2},
        )
        requests_mock.assert_has_calls(
            [
                call.__bool__(),
                call(*args1, **kwargs1),
                call().raise_for_status(),
                call().json(),
                call.__bool__(),
                call(*args2, **kwargs2),
                call().raise_for_status(),
                call().json(),
            ]
        )

        self.assertEqual(len(artifacts), 120)
        self.assertEqual(artifacts[0]["name"], "Foo-0")
        self.assertEqual(artifacts[119]["name"], "Foo-119")

    @patch("pontos.github.api.api.httpx.get")
    def test_get_repository_artifact(self, requests_mock: MagicMock):
        response = MagicMock(autospec=httpx.Response)
        response.json.return_value = {
            "id": 123,
            "name": "Foo",
        }

        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        artifacts = api.get_repository_artifact("foo/bar", "123")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/actions/artifacts/123",
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

        self.assertEqual(artifacts["id"], 123)
        self.assertEqual(artifacts["name"], "Foo")

    @patch("pontos.github.api.api.httpx.get")
    def test_get_repository_artifact_invalid(self, requests_mock: MagicMock):
        response = MagicMock(autospec=httpx.Response)
        response.is_success = False
        response.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Testing Status Message", request=None, response=response
        )

        requests_mock.return_value = response
        api = GitHubRESTApi("12345")

        with self.assertRaises(httpx.HTTPStatusError):
            api.get_repository_artifact("foo/bar", "123")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/actions/artifacts/123",
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.helper.Path")
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_repository_artifact(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        api = GitHubRESTApi("12345")
        download_file = path_mock()
        with api.download_repository_artifact(
            "foo/bar", "123", download_file
        ) as download_progress:
            args, kwargs = default_request(
                "GET",
                "https://api.github.com/repos/foo/bar/actions/artifacts/123/zip",  # pylint: disable=line-too-long
                timeout=DEFAULT_TIMEOUT,
            )
            requests_mock.assert_called_once_with(*args, **kwargs)
            response_headers.get.assert_called_once_with("content-length")

            self.assertIsNone(download_progress.length)

            it = iter(download_progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)

            with self.assertRaises(StopIteration):
                next(it)

    @patch("pontos.github.api.api.httpx.get")
    def test_get_workflow_run_artifacts(self, requests_mock: MagicMock):
        response = MagicMock()
        response.links = None
        response.json.return_value = {
            "total_count": 1,
            "artifacts": [
                {
                    "id": 11,
                    "node_id": "MDg6QXJ0aWZhY3QxMQ==",
                    "name": "Foo",
                }
            ],
        }
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        artifacts = api.get_workflow_run_artifacts("foo/bar", "123")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/actions/runs/123/artifacts",
            params={"per_page": 100, "page": 1},
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

        self.assertEqual(len(artifacts), 1)
        self.assertEqual(artifacts[0]["name"], "Foo")

    @patch("pontos.github.api.api.httpx.get")
    def test_get_workflow_run_artifacts_with_pagination(
        self, requests_mock: MagicMock
    ):
        response = MagicMock()
        response.links = None
        response.json.side_effect = [
            {
                "total_count": 120,
                "artifacts": [
                    {
                        "id": id,
                        "name": f"Foo-{id}",
                    }
                    for id in range(0, 100)
                ],
            },
            {
                "total_count": 120,
                "artifacts": [
                    {
                        "id": id,
                        "name": f"Foo-{id}",
                    }
                    for id in range(100, 120)
                ],
            },
        ]
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        artifacts = api.get_workflow_run_artifacts("foo/bar", "123")

        args1, kwargs1 = default_request(
            "https://api.github.com/repos/foo/bar/actions/runs/123/artifacts",
            params={"per_page": 100, "page": 1},
        )
        args2, kwargs2 = default_request(
            "https://api.github.com/repos/foo/bar/actions/runs/123/artifacts",
            params={"per_page": 100, "page": 2},
        )
        requests_mock.assert_has_calls(
            [
                call.__bool__(),
                call(*args1, **kwargs1),
                call().raise_for_status(),
                call().json(),
                call.__bool__(),
                call(*args2, **kwargs2),
                call().raise_for_status(),
                call().json(),
            ]
        )

        self.assertEqual(len(artifacts), 120)
        self.assertEqual(artifacts[0]["name"], "Foo-0")
        self.assertEqual(artifacts[119]["name"], "Foo-119")

    @patch("pontos.github.api.api.httpx.delete")
    def test_delete_repository_artifact(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.delete_repository_artifact("foo/bar", "123")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/actions/artifacts/123",
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.github.api.api.httpx.delete")
    def test_delete_repository_artifact_failure(self, requests_mock: MagicMock):
        response = MagicMock(autospec=httpx.Response)
        response.is_success = False
        response.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Delete Failed", request=None, response=response
        )

        requests_mock.return_value = response
        api = GitHubRESTApi("12345")

        with self.assertRaises(httpx.HTTPStatusError):
            api.delete_repository_artifact("foo/bar", "123")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/actions/artifacts/123",
        )
        requests_mock.assert_called_once_with(*args, **kwargs)



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
from datetime import date
from typing import List, Tuple


class ChangelogError(Exception):
    """
    Some error has occurred during changelog handling
    """


__UNRELEASED_FOOTER_MATCHER = re.compile("unreleased", re.IGNORECASE)
__UNRELEASED_HEADER_MATCHER = re.compile(
    r"[\(\[]{1}[0-9\.]*.*?[Uu]nreleased[\)\]]{1}"
)
__MASTER_MATCHER = re.compile("master|main|HEAD")

__UNRELEASED_SKELETON = """## [Unreleased]
### Added
### Changed
### Deprecated
### Removed
### Fixed

[Unreleased]: https://github.com/{}/{}/compare/{}...HEAD


"""


def add_skeleton(
    markdown: str,
    new_version: str,
    project_name: str,
    git_tag_prefix: str = "v",
    git_space: str = "greenbone",
) -> str:
    git_tag = f"{git_tag_prefix}{new_version}"
    tokens = _tokenize(markdown)
    updated_markdown = ""

    for tt, _, tc in tokens:
        if tt == "heading" and new_version in tc:
            prepared_skeleton = __UNRELEASED_SKELETON.format(
                git_space, project_name, git_tag
            )
            updated_markdown += prepared_skeleton + tc
        else:
            updated_markdown += tc

    return updated_markdown


def update(
    markdown: str,
    new_version: str,
    git_tag_prefix: str = "v",
    containing_version: str = None,
) -> Tuple[str, str]:
    """
    update tokenizes CHANGELOG.md and if a version is given it changes
    unreleased headline and link to given version.

    returns updated markdown and change log for further processing.
    """

    git_tag = f"{git_tag_prefix}{new_version}"
    tokens = _tokenize(markdown)
    unreleased_heading_count = 0
    changelog = ""
    updated_markdown = ""
    current_state = []
    previous_state = []
    unreleased = []

    for tt, hc, tc in tokens:
        previous_state = current_state.copy()

        if tt == "unreleased":
            if (
                containing_version and containing_version in tc
            ) or not containing_version:
                unreleased_heading_count = hc
                current_state = [tt]
        elif (
            tt == "heading"
            and unreleased_heading_count > 0
            and hc <= unreleased_heading_count
        ):
            current_state = []

        if "unreleased" not in current_state and "unreleased" in previous_state:
            changelog = _prepare_changelog(unreleased, new_version, git_tag)
            updated_markdown += changelog

        if "unreleased" not in current_state:
            updated_markdown += tc

        if "unreleased" in current_state:
            unreleased.append((tt, hc, tc))

    if "unreleased" in current_state and unreleased:
        changelog = _prepare_changelog(unreleased, new_version, git_tag)
        updated_markdown += changelog

    return (
        updated_markdown if changelog else "",
        changelog,
    )


def _prepare_changelog(
    tokens: List[Tuple[str, int, str]], new_version: str, git_tag: str
) -> str:
    current = ""
    previous = ""
    output = ""
    keyword_text = ""
    unreleased_link = ""

    for tt, _, tc in tokens:
        previous = current

        if tt == "unreleased":
            if new_version:
                tc = __UNRELEASED_HEADER_MATCHER.sub(f"[{new_version}]", tc)
                tc += f" - {date.today().isoformat()}"
            output += tc
        elif tt == "unreleased_link":
            if new_version:
                tc = __UNRELEASED_FOOTER_MATCHER.sub(new_version, tc)
                tc = __MASTER_MATCHER.sub(git_tag, tc)
            unreleased_link += tc + "\n\n"
        elif "kw_" in tt:
            if keyword_text.strip().count("\n") > 0:
                output += keyword_text
            keyword_text = tc
            current = tt
        elif "kw_" in previous:
            keyword_text += tc
        else:
            output += tc

    if keyword_text.strip().count("\n") > 0:
        output += keyword_text.strip() + "\n\n"

    output += unreleased_link

    return output


def __build_scanner():
    def token_handler(key: str):
        """
        generates a lambda for the regex scanner with a given key.

        This lambda will return a tuple: key, count # of token and token.

        The count is used to identify the level of heading on a special
        ended which can be used to identify when this section ended.
        """
        return lambda _, token: (key, token.count("#"), token)

    return re.Scanner(
        [
            (r"#{1,} [Aa]dded", token_handler("kw_added")),
            (r"#{1,} [Cc]hanged", token_handler("kw_changed")),
            (r"#{1,} [Dd]eprecated", token_handler("kw_deprecated")),
            (r"#{1,} [Rr]emoved", token_handler("kw_removed")),
            (r"#{1,} [Ff]ixed", token_handler("kw_fixed")),
            (r"#{1,} [Ss]ecurity", token_handler("kw_security")),
            (r"#{1,}.*(?=[Uu]nreleased).*", token_handler("unreleased")),
            (
                r"\[[Uu]nreleased\].*",
                token_handler("unreleased_link"),
            ),
            (r"#{1,} .*", token_handler("heading")),
            (r"\n", token_handler("newline")),
            (r"..*", token_handler("any")),
        ]
    )


__CHANGELOG_SCANNER = __build_scanner()


def _tokenize(
    markdown: str,
) -> List[Tuple[int, str, int, str],]:
    toks, remainder = __CHANGELOG_SCANNER.scan(markdown)
    if remainder != "":
        raise ChangelogError(f"unrecognized tokens in markdown: {remainder}")

    return toks



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos.github.api import GitHubRESTApi
from pontos.github.api.helper import FileStatus
from tests.github.api import default_request

here = Path(__file__).parent


class GitHubPullRequestsTestCase(unittest.TestCase):
    @patch("pontos.github.api.api.httpx.get")
    def test_pull_request_commits(self, requests_mock: MagicMock):
        response = MagicMock()
        response.links = None
        response.json.return_value = [{"sha": "1"}]
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        commits = api.pull_request_commits("foo/bar", pull_request=1)

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/pulls/1/commits",
            params={"per_page": "100"},
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

        self.assertEqual(len(commits), 1)
        self.assertEqual(commits[0]["sha"], "1")

    @patch("pontos.github.api.api.httpx.post")
    def test_create_pull_request(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.create_pull_request(
            "foo/bar",
            head_branch="foo",
            base_branch="main",
            title="Foo",
            body="This is bar",
        )

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/pulls",
            json={
                "head": "foo",
                "base": "main",
                "title": "Foo",
                "body": "This is bar",
            },
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.github.api.api.httpx.post")
    def test_update_pull_request(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.update_pull_request(
            "foo/bar",
            123,
            base_branch="main",
            title="Foo",
            body="This is bar",
        )

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/pulls/123",
            json={
                "base": "main",
                "title": "Foo",
                "body": "This is bar",
            },
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.github.api.api.httpx.post")
    def test_add_pull_request_comment(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.add_pull_request_comment(
            "foo/bar", pull_request=123, comment="This is a comment"
        )

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/issues/123/comments",
            json={"body": "This is a comment"},
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.github.api.api.httpx.get")
    def test_modified_files_in_pr(self, requests_mock: MagicMock):
        response = MagicMock()
        response.links = None
        response.json.return_value = json.loads(
            (here / "pr-files.json").read_text(encoding="utf-8")
        )
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        files = api.pull_request_files(
            "foo/bar", pull_request=1, status_list=[FileStatus.MODIFIED]
        )

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/pulls/1/files",
            params={"per_page": "100"},
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

        self.assertEqual(
            files,
            {
                FileStatus.MODIFIED: [
                    Path("gvm/protocols/gmpv2110/__init__.py"),
                    Path("tests/protocols/gmpv2110/entities/test_users.py"),
                    Path("tests/protocols/gmpv2110/entities/users/__init__.py"),
                    Path(
                        "tests/protocols/gmpv2110/"
                        "entities/users/test_modify_user.py"
                    ),
                ]
            },
        )

    @patch("pontos.github.api.api.httpx.get")
    def test_added_files_in_pr(self, requests_mock: MagicMock):
        response = MagicMock()
        response.links = None
        response.json.return_value = json.loads(
            (here / "pr-files.json").read_text(encoding="utf-8")
        )
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        files = api.pull_request_files(
            "foo/bar", pull_request=1, status_list=[FileStatus.ADDED]
        )

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/pulls/1/files",
            params={"per_page": "100"},
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

        self.assertEqual(
            files,
            {
                FileStatus.ADDED: [
                    Path("gvm/protocols/gmpv2110/entities/users.py"),
                    Path(
                        "tests/protocols/gmpv2110/entities/"
                        "users/test_create_user.py"
                    ),
                ]
            },
        )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Iterable, Optional

import httpx

from pontos.github.api.helper import JSON_OBJECT


class GitHubRESTWorkflowsMixin:
    def get_workflows(self, repo: str) -> Iterable[JSON_OBJECT]:
        """
        List all workflows of a repository

        Args:
            repo: GitHub repository (owner/name) to use

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the workflows as an iterable of dicts
        """
        api = f"/repos/{repo}/actions/workflows"
        return self._get_paged_items(api, "workflows")

    def get_workflow(self, repo: str, workflow: str) -> JSON_OBJECT:
        """
        Get the information for the given workflow

        Args:
            repo: GitHub repository (owner/name) to use
            workflow: ID of the workflow or workflow file name. For example
                `main.yml`.

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the workflows as a dict
        """
        api = f"/repos/{repo}/actions/workflows/{workflow}"
        response = self._request(api, request=httpx.get)
        response.raise_for_status()
        return response.json()

    def create_workflow_dispatch(
        self,
        repo: str,
        workflow: str,
        *,
        ref: str,
        inputs: Dict[str, str] = None,
    ):
        """
        Create a workflow dispatch event to manually trigger a GitHub Actions
        workflow run.

        Args:
            repo: GitHub repository (owner/name) to use
            workflow: ID of the workflow or workflow file name. For example
                `main.yml`.
            ref: The git reference for the workflow. The reference can be a
                branch or tag name.
            inputs: Input keys and values configured in the workflow file. Any
                default properties configured in the workflow file will be used
                when inputs are omitted.
        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Example:
            .. code-block:: python

            api = GitHubRESTApi("...")
            api.create_workflow_dispatch("foo/bar", "ci.yml", ref="main")
        """
        api = f"/repos/{repo}/actions/workflows/{workflow}/dispatches"
        data = {"ref": ref}

        if inputs:
            data["inputs"] = inputs

        response = self._request(api, data=data, request=httpx.post)
        response.raise_for_status()

    def get_workflow_runs(
        self,
        repo: str,
        workflow: Optional[str] = None,
        *,
        actor: Optional[str] = None,
        branch: Optional[str] = None,
        event: Optional[str] = None,
        status: Optional[str] = None,
        created: Optional[str] = None,
        exclude_pull_requests: Optional[bool] = None,
    ) -> Iterable[JSON_OBJECT]:
        # pylint: disable=line-too-long
        """
        List all workflow runs of a repository or of a specific workflow.

        Args:
            repo: GitHub repository (owner/name) to use
            workflow: ID of the workflow or workflow file name. For example
                `main.yml`.
            actor: Only return workflow runs of this user ID.
            branch: Only return workflow runs for a specific branch.
            event: Only returns workflows runs triggered by the event specified.
                For example, `push`, `pull_request` or `issue`.
                For more information, see https://docs.github.com/en/actions/using-workflows/events-that-trigger-workflows.
            status: Only return workflow runs with the check run status or
                conclusion that specified. For example, a conclusion can be
                `success` or a status can be `in_progress`. Can be one of:
                `completed`, `action_required`, `cancelled`, `failure`,
                `neutral`, `skipped`, `stale`, `success`, `timed_out`,
                `in_progress`, `queued`, `requested`, `waiting`.
            created: Only returns workflow runs created within the given
                date-time range. For more information on the syntax, see
                https://docs.github.com/en/search-github/getting-started-with-searching-on-github/understanding-the-search-syntax#query-for-dates
            exclude_pull_requests: If true pull requests are omitted from the
                response.

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the workflow runs as an iterable of dicts
        """

        api = (
            f"/repos/{repo}/actions/workflows/{workflow}/runs"
            if workflow
            else f"/repos/{repo}/actions/runs"
        )
        params = {}
        if actor:
            params["actor"] = actor
        if branch:
            params["branch"] = branch
        if event:
            params["event"] = event
        if status:
            params["status"] = status
        if created:
            params["created"] = created
        if exclude_pull_requests is not None:
            params["exclude_pull_requests"] = exclude_pull_requests

        return self._get_paged_items(api, "workflow_runs", params=params)

    def get_workflow_run(self, repo: str, run: str) -> JSON_OBJECT:
        """
        Get information about a single workflow run

        Args:
            repo: GitHub repository (owner/name) to use
            run: The ID of the workflow run

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the workflow runs as a dict
        """
        api = f"/repos/{repo}/actions/runs/{run}"
        response = self._request(api, request=httpx.get)
        response.raise_for_status()
        return response.json()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=no-member

import io
import unittest
from argparse import Namespace
from contextlib import redirect_stderr
from pathlib import Path
from unittest.mock import Mock

from pontos.github.api import FileStatus
from pontos.github.argparser import parse_args
from pontos.github.cmds import (
    create_pull_request,
    file_status,
    update_pull_request,
)


class TestArgparsing(unittest.TestCase):
    def setUp(self):
        self.term = Mock()

    def test_create_pr_parse_args(self):
        argv = [
            "pr",
            "create",
            "foo/bar",
            "baz",
            "main",
            "baz in main",
        ]

        parsed_args = parse_args(argv)

        template = Path().cwd() / "pontos/github/pr_template.md"

        self.assertEqual(parsed_args.command, "pr")
        self.assertEqual(parsed_args.token, "GITHUB_TOKEN")
        self.assertEqual(parsed_args.body, template.read_text(encoding="utf-8"))
        self.assertEqual(parsed_args.pr_func, create_pull_request)
        self.assertEqual(parsed_args.repo, "foo/bar")
        self.assertEqual(parsed_args.target, "main")
        self.assertEqual(parsed_args.title, "baz in main")

    def test_create_pr_parse_args_fail(self):
        argv = ["pr", "create", "foo/bar"]

        with self.assertRaises(SystemExit), redirect_stderr(io.StringIO()):
            parse_args(argv)

    def test_update_pr_parse_args(self):
        argv = [
            "-q",
            "--log-file",
            "foo",
            "pr",
            "update",
            "foo/bar",
            "123",
            "--body",
            "foo",
            "--target",
            "main",
            "--title",
            "baz in main",
        ]

        parsed_args = parse_args(argv)

        self.assertEqual(parsed_args.command, "pr")
        self.assertEqual(parsed_args.token, "GITHUB_TOKEN")
        self.assertEqual(parsed_args.body, "foo")
        self.assertEqual(parsed_args.pr_func, update_pull_request)
        self.assertEqual(parsed_args.repo, "foo/bar")
        self.assertEqual(parsed_args.pull_request, 123)
        self.assertEqual(parsed_args.target, "main")
        self.assertEqual(parsed_args.title, "baz in main")
        self.assertTrue(parsed_args.quiet)
        self.assertEqual(parsed_args.log_file, "foo")

    def test_update_pr_parse_args_fail(self):
        argv = ["pr", "update", "foo/bar"]

        with self.assertRaises(SystemExit), redirect_stderr(io.StringIO()):
            parse_args(argv)

    def test_fs_parse_args(self):
        argv = [
            "FS",
            "foo/bar",
            "8",
            "-o",
            "some.file",
        ]

        parsed_args = parse_args(argv)
        output = io.open(Path("some.file"), mode="w", encoding="utf-8")

        expected_args = Namespace(
            command="FS",
            func=file_status,
            repo="foo/bar",
            pull_request=8,
            output=output,
            status=[FileStatus.ADDED, FileStatus.MODIFIED],
            token="GITHUB_TOKEN",
        )

        self.assertEqual(type(parsed_args.output), type(expected_args.output))
        self.assertEqual(parsed_args.command, expected_args.command)
        self.assertEqual(parsed_args.func, expected_args.func)
        self.assertEqual(parsed_args.repo, expected_args.repo)
        self.assertEqual(parsed_args.pull_request, expected_args.pull_request)
        self.assertEqual(parsed_args.status, expected_args.status)
        self.assertEqual(parsed_args.token, expected_args.token)

        output.close()
        parsed_args.output.close()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=no-member

import io
import unittest
from argparse import Namespace
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos.github.api import FileStatus
from pontos.github.cmds import file_status

here = Path(__file__).parent


class TestArgparsing(unittest.TestCase):
    @patch("pontos.github.cmds.GitHubRESTApi")
    def test_file_status(self, api_mock):
        terminal = MagicMock()
        api_mock.return_value.pull_request_exists.return_value = True
        api_mock.return_value.pull_request_files.return_value = {
            FileStatus.ADDED: [Path("tests/github/foo/bar")],
            FileStatus.MODIFIED: [
                Path("tests/github/bar/baz"),
                Path("tests/github/baz/boo"),
            ],
        }
        test_file = Path("some.file")
        output = io.open(test_file, mode="w", encoding="utf-8")

        args = Namespace(
            command="FS",
            func=file_status,
            repo="foo/bar",
            pull_request=8,
            output=output,
            status=[FileStatus.ADDED, FileStatus.MODIFIED],
            token="GITHUB_TOKEN",
        )

        file_status(terminal, args)

        output.close()

        content = test_file.read_text(encoding="utf-8")
        self.assertEqual(
            content, f"{here}/foo/bar\n{here}/bar/baz\n{here}/baz/boo\n"
        )

        test_file.unlink()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import ContextManager, Iterable, Union

import httpx

from pontos.github.api.helper import JSON_OBJECT
from pontos.helper import DownloadProgressIterable, download


class GitHubRESTArtifactsMixin:
    def get_repository_artifacts(self, repo: str) -> Iterable[JSON_OBJECT]:
        """
        List all artifacts of a repository

        Args:
            repo: GitHub repository (owner/name) to use

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the artifacts in the repository as a dict
        """
        api = f"/repos/{repo}/actions/artifacts"
        return self._get_paged_items(api, "artifacts")

    def get_repository_artifact(self, repo: str, artifact: str) -> JSON_OBJECT:
        """
        Get a single artifact of a repository

        Args:
            repo: GitHub repository (owner/name) to use
            artifact: ID of the artifact

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the artifact as a dict
        """
        api = f"/repos/{repo}/actions/artifacts/{artifact}"
        response: httpx.Response = self._request(api, request=httpx.get)
        response.raise_for_status()
        return response.json()

    def download_repository_artifact(
        self, repo: str, artifact: str, destination: Union[Path, str]
    ) -> ContextManager[DownloadProgressIterable]:
        """
        Download a repository artifact zip file

        Args:
            repo: GitHub repository (owner/name) to use
            artifact: ID of the artifact
            destination: A path where to store the downloaded file

        Raises:
            HTTPError if the request was invalid

        Example:
            .. code-block:: python

            api = GitHubRESTApi("...")

            print("Downloading", end="")

            with api.download_repository_artifact(
                "org/repo", "123", "/tmp/artifact.zip"
            ) as progress_iterator:
                for progress in progress_iterator:
                    print(".", end="")
        """
        api = f"{self.url}/repos/{repo}/actions/artifacts/{artifact}/zip"
        return download(api, destination, headers=self._request_headers())

    def get_workflow_run_artifacts(
        self, repo: str, run: str
    ) -> Iterable[JSON_OBJECT]:
        """
        List all artifacts for a workflow run

        Args:
            repo: GitHub repository (owner/name) to use
            run: The unique identifier of the workflow run

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the artifacts in the workflow as a dict
        """
        api = f"/repos/{repo}/actions/runs/{run}/artifacts"
        return self._get_paged_items(api, "artifacts")

    def delete_repository_artifact(self, repo: str, artifact: str):
        """
        Delete an artifact of a repository

        Args:
            repo: GitHub repository (owner/name) to use
            artifact: ID of the artifact

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.
        """
        api = f"/repos/{repo}/actions/artifacts/{artifact}"
        response: httpx.Response = self._request(api, request=httpx.delete)
        response.raise_for_status()



# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, Iterator, Optional, Union

import httpx

DEFAULT_TIMEOUT = 1000
DEFAULT_CHUNK_SIZE = 4096


class DownloadProgressIterable:
    def __init__(
        self,
        *,
        content_iterator: Iterator,
        url: str,
        destination: Path,
        length: int,
    ):
        self._content_iterator = content_iterator
        self._url = url
        self._destination = destination
        self._length = None if length is None else int(length)

    @property
    def length(self) -> Optional[int]:
        """
        Size in bytes of the to be downloaded file or None if the size is not
        available
        """
        return self._length

    @property
    def destination(self) -> Path:
        """
        Destination path of the to be downloaded file
        """
        return self._destination

    @property
    def url(self) -> str:
        return self._url

    def _download(self) -> Iterator[Optional[float]]:
        dl = 0
        with self._destination.open("wb") as f:
            for content in self._content_iterator:
                dl += len(content)
                f.write(content)
                yield dl / self._length if self._length else None

    def __iter__(self) -> Iterator[Optional[float]]:
        return self._download()

    def run(self):
        """
        Just run the download without caring about the progress
        """
        try:
            it = iter(self)
            while True:
                next(it)
        except StopIteration:
            pass


@contextmanager
def download(
    url: str,
    destination: Optional[Union[Path, str]] = None,
    *,
    headers: Dict[str, Any] = None,
    params: Dict[str, Any] = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    timeout: int = DEFAULT_TIMEOUT,
) -> Generator[DownloadProgressIterable, None, None]:
    """Download file in url to filename

    Arguments:
        url: The url of the file we want to download
        destination: Path of the file to store the download in. If set it will
                     be derived from the passed URL.
        headers: HTTP headers to use for the download
        params: HTTP request parameters to use for the download
        chunk_size: Download file in chunks of this size
        timeout: Connection timeout

    Raises:
        HTTPError if the request was invalid

    Returns:
        A DownloadProgressIterator that yields the progress of the download in
        percent for each downloaded chunk or None for each chunk if the progress
        is unknown.

    Example:
        .. code-block:: python

        with download("https://example.com/some/file")) as progress_it:
            for progress in progress_it:
                print(progress)
    """
    destination = (
        Path(url.split("/")[-1]) if not destination else Path(destination)
    )

    with httpx.stream(
        "GET",
        url,
        timeout=timeout,
        follow_redirects=True,
        headers=headers,
        params=params,
    ) as response:
        response.raise_for_status()

        total_length = response.headers.get("content-length")

        yield DownloadProgressIterable(
            url=url,
            content_iterator=response.iter_bytes(chunk_size=chunk_size),
            destination=destination,
            length=total_length,
        )


def shell_cmd_runner(args: Iterable[str]) -> subprocess.CompletedProcess:
    return subprocess.run(
        args,
        shell=True,
        check=True,
        errors="utf-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import unittest
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from unittest.mock import MagicMock, call, patch

import httpx

from pontos import changelog, release
from pontos.release.helper import version


class ReleaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ["GITHUB_TOKEN"] = "foo"
        os.environ["GITHUB_USER"] = "bar"
        self.valid_gh_release_response = (
            '{"zipball_url": "zip", "tarball_url":'
            ' "tar", "upload_url":"upload"}'
        )

    @patch("pontos.release.release.shell_cmd_runner")
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_release_successfully(
        self,
        _changelog_mock,
        _version_mock,
        _requests_mock,
        _requests2_mock,
        _path_mock,
        _shell_mock,
    ):
        _version_mock.main.return_value = (True, "MyProject.conf")
        _changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        _requests_mock.post.return_value = fake_post
        _requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "--release-version",
            "0.0.1",
            "--next-version",
            "0.0.2dev",
        ]

        with redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)

    @patch("pontos.release.release.shell_cmd_runner")
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_release_conventional_commits_successfully(
        self,
        _changelog_mock,
        _version_mock,
        _requests_mock,
        _requests2_mock,
        _path_mock,
        _shell_mock,
    ):
        _version_mock.main.return_value = (True, "MyProject.conf")
        _changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        _requests_mock.post.return_value = fake_post
        _requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "-CC",
        ]

        with redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)

    @patch("pontos.release.release.shell_cmd_runner")
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_not_release_successfully_when_github_create_release_fails(
        self,
        _changelog_mock,
        _version_mock,
        _requests_mock,
        _requests2_mock,
        _path_mock,
        _shell_mock,
    ):

        _version_mock.main.return_value = (True, "MyProject.conf")
        _changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 401
        fake_post.text = self.valid_gh_release_response
        fake_post.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Authorization required",
            response=fake_post,
            request=_requests_mock.post,
        )
        _requests_mock.post.return_value = fake_post
        _requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "--release-version",
            "0.0.1",
        ]

        with redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertFalse(released)

    @patch("pontos.release.release.shell_cmd_runner")
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_release_to_specific_git_remote(
        self,
        _changelog_mock,
        _version_mock,
        _requests_mock,
        _requests2_mock,
        _path_mock,
        shell_mock,
    ):
        _version_mock.main.return_value = (True, "MyProject.conf")
        _changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        _requests_mock.post.return_value = fake_post
        _requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "--project",
            "foo",
            "--release-version",
            "0.0.1",
            "--next-version",
            "0.0.2.dev1",
            "--git-remote-name",
            "upstream",
            "--git-signing-key",
            "1234",
        ]

        with redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)

        shell_mock.assert_has_calls(
            [
                call("git push --follow-tags upstream"),
                call("git add MyProject.conf"),
                call("git add *__version__.py || echo 'ignoring __version__'"),
                call("git add CHANGELOG.md"),
                call(
                    "git commit -S1234 --no-verify -m 'Automatic adjustments "
                    "after release\n\n"
                    "* Update to version 0.0.2.dev1\n"
                    "* Add empty changelog after 0.0.1'"
                ),
            ]
        )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import httpx

from pontos.github.api.helper import JSON, RepositoryType


class GitHubRESTOrganizationsMixin:
    def organisation_exists(self, orga: str) -> bool:
        """
        Check if an organization exists

        Args:
            repo: GitHub repository (owner/name) to use
        """
        api = f"/orgs/{orga}"
        response: httpx.Response = self._request(api)
        return response.is_success

    def get_organization_repository_number(
        self, orga: str, repository_type: RepositoryType
    ) -> int:
        """
        Get total number of repositories of organization

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number to check
        """
        api = f"/orgs/{orga}"
        response: httpx.Response = self._request(api)
        response.raise_for_status()
        response_json = response.json()

        if repository_type == RepositoryType.PUBLIC:
            return response_json["public_repos"]
        if repository_type == RepositoryType.PRIVATE:
            return response_json["total_private_repos"]
            # Use ALL for currently unsupported "types" (INTERNAL, FORKS ...)
        return (
            response_json["public_repos"] + response_json["total_private_repos"]
        )

    def get_repositories(
        self,
        orga: str,
        repository_type: RepositoryType = RepositoryType.ALL,
    ) -> JSON:
        """
        Get information about organization repositories

        Args:
            orga: GitHub organization to use
        """
        api = f"/orgs/{orga}/repos"

        count = self.get_organization_repository_number(
            orga=orga, repository_type=repository_type
        )
        downloaded = 0

        repos = []
        params = {"type": repository_type.value, "per_page": 100}
        page = 0
        while count - downloaded > 0:
            page += 1
            params["page"] = page
            response: httpx.Response = self._request(api, params=params)
            response.raise_for_status()

            repos.extend(response.json())
            downloaded = len(repos)

        return repos



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos.github.api import GitHubRESTApi
from tests.github.api import default_request

here = Path(__file__).parent


class GitHubContentTestCase(unittest.TestCase):
    @patch("pontos.github.api.api.httpx.get")
    def test_path_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.ok = True
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.path_exists(repo="foo/bar", path="baz", branch="main")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/contents/baz",
            params={
                "ref": "main",
            },
        )
        requests_mock.assert_called_once_with(*args, **kwargs)
        self.assertTrue(exists)

    @patch("pontos.github.api.api.httpx.get")
    def test_path_not_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.is_success = False
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.path_exists(repo="foo/bar", path="baz", branch="main")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/contents/baz",
            params={
                "ref": "main",
            },
        )
        requests_mock.assert_called_once_with(*args, **kwargs)
        self.assertFalse(exists)



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import io
import unittest
from pathlib import Path
from unittest.mock import MagicMock

from pontos.version.helper import VersionError
from pontos.version.version import VersionCommand


class VersionCommandTestCase(unittest.TestCase):
    def test_missing_cmd(self):
        with contextlib.redirect_stdout(io.StringIO()) as buf:
            VersionCommand().run()
            self.assertEqual(
                buf.getvalue(),
                "usage: version [-h] [--quiet] {verify,show,update} ...\n",
            )

    def test_missing_file(self):
        with self.assertRaises(
            VersionError, msg="Could not find whatever file."
        ):
            VersionCommand(project_file_path=Path("whatever")).run(
                args=["show"]
            )

    def test_print_current_version(self):
        print_mock = MagicMock()
        VersionCommand.get_current_version = MagicMock(return_value="1.2.3")
        VersionCommand._print = print_mock  # pylint: disable=protected-access

        cmd = VersionCommand()
        cmd.print_current_version()

        print_mock.assert_called_with("1.2.3")



# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from argparse import Namespace
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from subprocess import CompletedProcess
from unittest.mock import MagicMock, call, patch

from pontos import changelog
from pontos.changelog.conventional_commits import parse_args


@dataclass
class StdOutput:
    stdout: bytes


class ConventionalCommitsTestCase(unittest.TestCase):
    @patch("pontos.changelog.conventional_commits.shell_cmd_runner")
    def test_changelog_builder(self, shell_mock):
        terminal = MagicMock()

        today = datetime.today().strftime("%Y-%m-%d")

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"

        shell_mock.side_effect = [
            CompletedProcess(args=None, returncode=1, stdout="v0.0.1"),
            CompletedProcess(
                args=None,
                returncode=1,
                stdout=(
                    "1234567 Add: foo bar\n"
                    "8abcdef Add: bar baz\n"
                    "8abcd3f Add bar baz\n"
                    "8abcd3d Adding bar baz\n"
                    "1337abc Change: bar to baz\n"
                    "42a42a4 Remove: foo bar again\n"
                    "fedcba8 Test: bar baz testing\n"
                    "dead901 Refactor: bar baz ref\n"
                    "fedcba8 Fix: bar baz fixing\n"
                    "d0c4d0c Doc: bar baz documenting\n"
                ),
            ),
        ]

        cargs = Namespace(
            current_version="0.0.1",
            next_version=release_version,
            output=output,
            space="foo",
            project="bar",
            config=config_toml,
        )
        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            args=cargs,
        )
        expected_output = f"""# Changelog

All notable changes to this project will be documented in this file.

## [0.0.2] - {today}

## Added
* foo bar [1234567](https://github.com/foo/bar/commit/1234567)
* bar baz [8abcdef](https://github.com/foo/bar/commit/8abcdef)

## Removed
* foo bar again [42a42a4](https://github.com/foo/bar/commit/42a42a4)

## Changed
* bar to baz [1337abc](https://github.com/foo/bar/commit/1337abc)

## Bug Fixes
* bar baz fixing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

## Documentation
* bar baz documenting [d0c4d0c](https://github.com/foo/bar/commit/d0c4d0c)

## Refactor
* bar baz ref [dead901](https://github.com/foo/bar/commit/dead901)

## Testing
* bar baz testing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

[0.0.2]: https://github.com/foo/bar/compare/0.0.1...0.0.2"""
        output_file = changelog_builder.create_changelog_file()
        self.assertEqual(
            output_file.read_text(encoding="utf-8"), expected_output
        )

        shell_mock.assert_has_calls(
            [
                call("git tag | sort -V | tail -1"),
                call('git log "v0.0.1..HEAD" --oneline'),
            ]
        )

        output_file.unlink()

    @patch("pontos.changelog.conventional_commits.shell_cmd_runner")
    def test_changelog_builder_no_commits(self, shell_mock):
        terminal = MagicMock()

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"

        shell_mock.return_value = CompletedProcess(
            args=None, returncode=1, stdout="1234567 foo bar\n8abcdef bar baz\n"
        )

        cargs = Namespace(
            current_version="0.0.1",
            next_version=release_version,
            output=output,
            space="foo",
            project="bar",
            config=config_toml,
        )
        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            args=cargs,
        )

        with self.assertRaises(SystemExit):
            output_file = changelog_builder.create_changelog_file()

            self.assertIsNone(output_file)
            shell_mock.assert_called_with(
                'git log "$(git describe --tags --abbrev=0)..HEAD" --oneline',
            )

    @patch("pontos.changelog.conventional_commits.shell_cmd_runner")
    def test_changelog_builder_no_tag(self, shell_mock):
        terminal = MagicMock()

        today = datetime.today().strftime("%Y-%m-%d")

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"
        shell_mock.side_effect = [
            CompletedProcess(args=None, returncode=1, stdout=""),
            CompletedProcess(
                args=None,
                returncode=1,
                stdout=(
                    "1234567 Add: foo bar\n"
                    "8abcdef Add: bar baz\n"
                    "8abcd3f Add bar baz\n"
                    "8abcd3d Adding bar baz\n"
                    "1337abc Change: bar to baz\n"
                    "42a42a4 Remove: foo bar again\n"
                    "fedcba8 Test: bar baz testing\n"
                    "dead901 Refactor: bar baz ref\n"
                    "fedcba8 Fix: bar baz fixing\n"
                    "d0c4d0c Doc: bar baz documenting\n"
                ),
            ),
        ]

        cargs = Namespace(
            current_version="0.0.1",
            next_version=release_version,
            output=output,
            space="foo",
            project="bar",
            config=config_toml,
        )
        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            args=cargs,
        )
        expected_output = f"""# Changelog

All notable changes to this project will be documented in this file.

## [0.0.2] - {today}

## Added
* foo bar [1234567](https://github.com/foo/bar/commit/1234567)
* bar baz [8abcdef](https://github.com/foo/bar/commit/8abcdef)

## Removed
* foo bar again [42a42a4](https://github.com/foo/bar/commit/42a42a4)

## Changed
* bar to baz [1337abc](https://github.com/foo/bar/commit/1337abc)

## Bug Fixes
* bar baz fixing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

## Documentation
* bar baz documenting [d0c4d0c](https://github.com/foo/bar/commit/d0c4d0c)

## Refactor
* bar baz ref [dead901](https://github.com/foo/bar/commit/dead901)

## Testing
* bar baz testing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

[0.0.2]: https://github.com/foo/bar/compare/0.0.1...0.0.2"""
        output_file = changelog_builder.create_changelog_file()
        self.assertEqual(
            output_file.read_text(encoding="utf-8"), expected_output
        )

        shell_mock.assert_has_calls(
            [
                call("git tag | sort -V | tail -1"),
                call("git log HEAD --oneline"),
            ]
        )

        output_file.unlink()

    @patch("pontos.changelog.conventional_commits.shell_cmd_runner")
    def test_changelog_builder_no_conventional_commits(self, shell_mock):
        terminal = MagicMock()

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"
        shell_mock.return_value = CompletedProcess(
            args="foo", returncode=1, stdout=None
        )
        cargs = Namespace(
            current_version="0.0.1",
            next_version=release_version,
            output=output,
            space="foo",
            project="bar",
            config=config_toml,
        )
        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            args=cargs,
        )

        with self.assertRaises(SystemExit):
            output_file = changelog_builder.create_changelog_file()

            self.assertIsNone(output_file)
            shell_mock.assert_called_with(
                'git log "$(git describe --tags --abbrev=0)..HEAD" --oneline',
            )

    def test_parse_args(self):
        parsed_args = parse_args(
            ["-q", "--log-file", "blah", "--project", "urghs"]
        )

        self.assertTrue(parsed_args.quiet)
        self.assertEqual(parsed_args.log_file, "blah")
        self.assertEqual(parsed_args.project, "urghs")



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, Iterable


class PullRequestState(Enum):
    OPEN = "open"
    CLOSED = "closed"


@dataclass
class Label:
    """
    A label of a pull request or issue
    """

    name: str


@dataclass
class Ref:
    """
    A git branch reference
    """

    name: str
    sha: str


@dataclass
class GitHubPullRequestEvent:
    """
    Event data of a GitHub Pull Request

    https://docs.github.com/en/developers/webhooks-and-events/webhooks/webhook-events-and-payloads#pull_request
    """

    draft: bool
    number: int
    labels: Iterable[str]
    title: str
    merged: bool
    state: PullRequestState
    base: Ref
    head: Ref

    def __init__(self, pull_request_data: Dict[str, str]):
        data = pull_request_data or {}

        self.draft = data.get("draft")
        self.number = data.get("number")
        self.labels = [Label(label.get("name")) for label in data.get("labels")]
        self.title = data.get("title")
        self.merged = data.get("merged")
        self.state = PullRequestState(data.get("state"))

        base = data.get("base") or {}
        self.base = Ref(base.get("ref"), base.get("sha"))

        head = data.get("head") or {}
        self.head = Ref(head.get("ref"), head.get("sha"))


@dataclass
class GitHubEvent:
    """
    GitHub Actions provides event data for the running action as JSON data in
    a local file at the runner.

    The JSON data for the events is specified at
    https://docs.github.com/en/developers/webhooks-and-events/webhooks/webhook-events-and-payloads
    """

    pull_request: GitHubPullRequestEvent

    def __init__(self, event_path: Path):
        content = event_path.read_text(encoding="utf-8")
        self._event_data = json.loads(content) if content else {}
        pull_request_data = self._event_data.get("pull_request")
        self.pull_request = (
            GitHubPullRequestEvent(pull_request_data)
            if pull_request_data
            else None
        )

    def __str__(self) -> str:
        return json.dumps(self._event_data, indent=2)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

import httpx

from pontos.github.api import GitHubRESTApi
from pontos.helper import DEFAULT_TIMEOUT
from tests.github.api import default_request

here = Path(__file__).parent


class GitHubReleaseTestCase(unittest.TestCase):
    @patch("pontos.github.api.api.httpx.post")
    def test_create_release(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.create_release(
            "foo/bar",
            "v1.2.3",
            name="Foo v1.2.3",
            body="This is a release",
        )

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/releases",
            json={
                "tag_name": "v1.2.3",
                "name": "Foo v1.2.3",
                "body": "This is a release",
                "draft": False,
                "prerelease": False,
            },
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.github.api.api.httpx.get")
    def test_release_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.ok = True
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.release_exists("foo/bar", "v1.2.3")

        args, kwargs = default_request(
            "https://api.github.com/repos/foo/bar/releases/tags/v1.2.3",
        )
        requests_mock.assert_called_once_with(*args, **kwargs)
        self.assertTrue(exists)

    @patch("pontos.github.api.api.httpx.get")
    def test_release(self, requests_mock: MagicMock):
        response = MagicMock()
        response.json.return_value = json.loads(
            (here / "release-response.json").read_text(encoding="utf-8")
        )

        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        data = api.release("greenbone/pontos", "v21.11.0")

        args, kwargs = default_request(
            "https://api.github.com/repos/greenbone/pontos/releases/tags/v21.11.0",  # pylint: disable=line-too-long
        )
        requests_mock.assert_called_once_with(*args, **kwargs)

        self.assertEqual(data["id"], 52499047)

    @patch("pontos.helper.Path")
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_release_tarball(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        api = GitHubRESTApi("12345")
        download_file = path_mock()
        with api.download_release_tarball(
            "greenbone/pontos", "v21.11.0", download_file
        ) as download_progress:
            args, kwargs = default_request(
                "GET",
                "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
                headers=None,
                timeout=DEFAULT_TIMEOUT,
            )
            requests_mock.assert_called_once_with(*args, **kwargs)
            response_headers.get.assert_called_once_with("content-length")

            self.assertIsNone(download_progress.length)

            it = iter(download_progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)

            with self.assertRaises(StopIteration):
                next(it)

    @patch("pontos.helper.Path")
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_release_tarball_with_content_length(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = "9"
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        api = GitHubRESTApi("12345")
        download_file = path_mock()
        with api.download_release_tarball(
            "greenbone/pontos", "v21.11.0", download_file
        ) as download_progress:
            args, kwargs = default_request(
                "GET",
                "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
                headers=None,
                timeout=DEFAULT_TIMEOUT,
            )
            requests_mock.assert_called_once_with(*args, **kwargs)
            response_headers.get.assert_called_once_with("content-length")

            self.assertEqual(download_progress.length, 9)

            it = iter(download_progress)
            progress = next(it)
            self.assertEqual(progress, 1 / 3)
            progress = next(it)
            self.assertEqual(progress, 2 / 3)
            progress = next(it)
            self.assertEqual(progress, 1)

            with self.assertRaises(StopIteration):
                next(it)

    @patch("pontos.helper.Path")
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_release_zip(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        api = GitHubRESTApi("12345")
        download_file = path_mock()
        with api.download_release_zip(
            "greenbone/pontos", "v21.11.0", download_file
        ) as download_progress:
            args, kwargs = default_request(
                "GET",
                "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.zip",  # pylint: disable=line-too-long
                headers=None,
                timeout=DEFAULT_TIMEOUT,
            )
            requests_mock.assert_called_once_with(*args, **kwargs)
            response_headers.get.assert_called_once_with("content-length")

            self.assertIsNone(download_progress.length)

            it = iter(download_progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)

            with self.assertRaises(StopIteration):
                next(it)

    @patch("pontos.helper.Path")
    @patch("pontos.github.api.api.httpx.get")
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_release_assets(
        self,
        stream_mock: MagicMock,
        request_mock: MagicMock,
        _path_mock: MagicMock,
    ):
        response = MagicMock()
        response.iter_bytes.side_effect = [
            [b"foo", b"bar", b"baz"],
            [b"lorem", b"ipsum"],
        ]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        stream_mock.return_value = response_stream

        response = MagicMock()
        response.json.side_effect = [
            {
                "assets_url": "https://api.github.com/repos/greenbone/pontos/releases/52499047/assets",  # pylint: disable=line-too-long
            },
            [
                {
                    "name": "foo-21.11.0.tar.gz",
                    "browser_download_url": "https://github.com/greenbone/pontos/releases/download/v21.11.0/foo-21.11.0.tar.gz",  # pylint: disable=line-too-long
                },
                {
                    "name": "bar-21.11.0.zip",
                    "browser_download_url": "https://github.com/greenbone/pontos/releases/download/v21.11.0/bar-21.11.0.zip",  # pylint: disable=line-too-long
                },
            ],
        ]

        request_mock.return_value = response

        api = GitHubRESTApi("12345")
        download_iter = iter(
            api.download_release_assets("greenbone/pontos", "v21.11.0")
        )

        download_progress = next(download_iter)
        self.assertIsNone(download_progress.length)

        progress_it = iter(download_progress)
        progress = next(progress_it)
        self.assertIsNone(progress)
        progress = next(progress_it)
        self.assertIsNone(progress)
        progress = next(progress_it)
        self.assertIsNone(progress)

        with self.assertRaises(StopIteration):
            next(progress_it)

        download_progress = next(download_iter)
        progress_it = iter(download_progress)
        progress = next(progress_it)
        self.assertIsNone(progress)
        progress = next(progress_it)

        with self.assertRaises(StopIteration):
            next(progress_it)

        with self.assertRaises(StopIteration):
            next(download_iter)

        args1, kwargs1 = default_request(
            "https://api.github.com/repos/greenbone/pontos/releases/tags/v21.11.0",  # pylint: disable=line-too-long
        )
        args2, kwargs2 = default_request(
            "https://api.github.com/repos/greenbone/pontos/releases/52499047/assets",  # pylint: disable=line-too-long
        )
        request_mock.assert_has_calls(
            [
                call(
                    *args1,
                    **kwargs1,
                ),
                call().raise_for_status(),
                call().json(),
                call(
                    *args2,
                    **kwargs2,
                ),
                call().raise_for_status(),
                call().json(),
            ]
        )

    @patch("pontos.github.api.release.Path")
    @patch("pontos.github.api.api.httpx.get")
    def test_download_release_assets_no_assets(
        self,
        request_mock: MagicMock,
        _path_mock: MagicMock,
    ):
        response = MagicMock()
        response.json.return_value = {}
        request_mock.return_value = response

        api = GitHubRESTApi("12345")
        download_iter = iter(
            api.download_release_assets("greenbone/pontos", "v21.11.0")
        )

        with self.assertRaises(StopIteration):
            next(download_iter)

    @patch("pontos.github.api.release.Path")
    @patch("pontos.github.api.api.httpx.get")
    def test_download_release_assets_no_files(
        self,
        request_mock: MagicMock,
        _path_mock: MagicMock,
    ):
        response = MagicMock()
        response.json.side_effect = [
            {
                "assets_url": "https://api.github.com/repos/greenbone/pontos/releases/52499047/assets",  # pylint: disable=line-too-long
            },
            [
                {
                    "name": "foo.txt",
                    "browser_download_url": "https://github.com/greenbone/pontos/releases/download/v21.11.0/foo.txt",  # pylint: disable=line-too-long
                },
                {
                    "name": "foo.txt.asc",
                    "browser_download_url": "https://github.com/greenbone/pontos/releases/download/v21.11.0/foo.txt.asc",  # pylint: disable=line-too-long
                },
            ],
        ]
        request_mock.return_value = response

        api = GitHubRESTApi("12345")
        download_iter = iter(
            api.download_release_assets("greenbone/pontos", "v21.11.0")
        )

        with self.assertRaises(StopIteration):
            next(download_iter)

    @patch("pontos.github.api.api.httpx.post")
    @patch("pontos.github.api.api.httpx.get")
    def test_upload_release_assets(
        self, get_mock: MagicMock, post_mock: MagicMock
    ):
        response = MagicMock()
        response.json.return_value = {
            "upload_url": "https://uploads.github.com/repos/greenbone/pontos/releases/52499047/assets{?name,label}",  # pylint: disable=line-too-long
        }
        get_mock.return_value = response

        file1 = MagicMock()
        file1.name = "foo.txt"
        content1 = b"foo"
        file1.read_bytes.return_value = content1
        file2 = MagicMock()
        file2.name = "bar.pdf"
        content2 = b"bar"
        file2.read_bytes.return_value = content2
        upload_files = [file1, file2]

        post_response = MagicMock()
        post_mock.return_value = post_response

        api = GitHubRESTApi("12345")
        upload_it = iter(
            api.upload_release_assets(
                "greenbone/pontos", "v21.11.0", upload_files
            )
        )
        args, kwargs = default_request(
            "https://uploads.github.com/repos/greenbone/pontos/releases/52499047/assets",  # pylint: disable=line-too-long
            headers={
                "Accept": "application/vnd.github.v3+json",
                "Authorization": "token 12345",
                "Content-Type": "application/octet-stream",
            },
            params={"name": "foo.txt"},
            content=content1,
        )
        uploaded_file = next(upload_it)
        post_mock.assert_called_with(*args, **kwargs)
        self.assertEqual(uploaded_file, file1)

        args, kwargs = default_request(
            "https://uploads.github.com/repos/greenbone/pontos/releases/52499047/assets",  # pylint: disable=line-too-long
            headers={
                "Accept": "application/vnd.github.v3+json",
                "Authorization": "token 12345",
                "Content-Type": "application/octet-stream",
            },
            params={"name": "bar.pdf"},
            content=content2,
        )
        uploaded_file = next(upload_it)
        post_mock.assert_called_with(*args, **kwargs)
        self.assertEqual(uploaded_file, file2)

        args, kwargs = default_request(
            "https://api.github.com/repos/greenbone/pontos/releases/tags/v21.11.0",  # pylint: disable=line-too-long
        )
        get_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.github.api.api.httpx.post")
    @patch("pontos.github.api.api.httpx.get")
    def test_upload_release_assets_no_release(
        self, get_mock: MagicMock, post_mock: MagicMock
    ):
        response = MagicMock()
        response.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Authentication required 401",
            request=MagicMock(),
            response=response,
        )
        get_mock.return_value = response

        file1 = MagicMock()
        file2 = MagicMock()
        upload_files = [file1, file2]

        post_response = MagicMock()
        post_mock.return_value = post_response

        api = GitHubRESTApi("12345")
        upload_it = iter(
            api.upload_release_assets(
                "greenbone/pontos", "v21.11.0", upload_files
            )
        )
        with self.assertRaises(httpx.HTTPError):
            next(upload_it)

        args, kwargs = default_request(
            "https://api.github.com/repos/greenbone/pontos/releases/tags/v21.11.0",  # pylint: disable=line-too-long
        )
        get_mock.assert_called_once_with(*args, **kwargs)

    @patch("pontos.github.api.api.httpx.post")
    @patch("pontos.github.api.api.httpx.get")
    def test_upload_release_assets_upload_fails(
        self, get_mock: MagicMock, post_mock: MagicMock
    ):
        response = MagicMock()
        response.json.return_value = {
            "upload_url": "https://uploads.github.com/repos/greenbone/pontos/releases/52499047/assets{?name,label}",  # pylint: disable=line-too-long
        }
        get_mock.return_value = response

        file1 = MagicMock()
        file1.name = "foo.txt"
        content1 = b"foo"
        file1.read_bytes.return_value = content1
        file2 = MagicMock()
        file2.name = "bar.pdf"
        content2 = b"bar"
        file2.read_bytes.return_value = content2
        upload_files = [file1, file2]

        post_response = MagicMock()
        post_response.raise_for_status.side_effect = [
            "",
            httpx.HTTPStatusError(
                "Internal Server Error",
                request=MagicMock(),
                response=response,
            ),
        ]
        post_mock.return_value = post_response

        api = GitHubRESTApi("12345")
        upload_it = iter(
            api.upload_release_assets(
                "greenbone/pontos", "v21.11.0", upload_files
            )
        )
        uploaded_file = next(upload_it)
        args, kwargs = default_request(
            "https://uploads.github.com/repos/greenbone/pontos/releases/52499047/assets",  # pylint: disable=line-too-long
            headers={
                "Accept": "application/vnd.github.v3+json",
                "Authorization": "token 12345",
                "Content-Type": "application/octet-stream",
            },
            params={"name": "foo.txt"},
            content=content1,
        )
        post_mock.assert_called_with(*args, **kwargs)
        self.assertEqual(uploaded_file, file1)

        with self.assertRaises(httpx.HTTPError):
            next(upload_it)

        args, kwargs = default_request(
            "https://api.github.com/repos/greenbone/pontos/releases/tags/v21.11.0",  # pylint: disable=line-too-long
        )
        get_mock.assert_called_once_with(*args, **kwargs)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable, Dict, Iterable, Iterator, Optional

import httpx

from pontos.github.api.artifacts import GitHubRESTArtifactsMixin
from pontos.github.api.branch import GitHubRESTBranchMixin
from pontos.github.api.contents import GitHubRESTContentMixin
from pontos.github.api.helper import (
    DEFAULT_GITHUB_API_URL,
    DEFAULT_TIMEOUT_CONFIG,
    JSON,
    JSON_OBJECT,
    _get_next_url,
)
from pontos.github.api.labels import GitHubRESTLabelsMixin
from pontos.github.api.organizations import GitHubRESTOrganizationsMixin
from pontos.github.api.pull_requests import GitHubRESTPullRequestsMixin
from pontos.github.api.release import GitHubRESTReleaseMixin
from pontos.github.api.workflows import GitHubRESTWorkflowsMixin


class GitHubRESTApi(
    GitHubRESTArtifactsMixin,
    GitHubRESTBranchMixin,
    GitHubRESTContentMixin,
    GitHubRESTLabelsMixin,
    GitHubRESTOrganizationsMixin,
    GitHubRESTPullRequestsMixin,
    GitHubRESTReleaseMixin,
    GitHubRESTWorkflowsMixin,
):
    """GitHubRESTApi Mixin"""

    def __init__(
        self,
        token: Optional[str] = None,
        url: Optional[str] = DEFAULT_GITHUB_API_URL,
        *,
        timeout: httpx.Timeout = DEFAULT_TIMEOUT_CONFIG,
    ) -> None:
        self.token = token
        self.url = url
        self.timeout = timeout

    def _request_headers(
        self, *, content_type: Optional[str] = None
    ) -> Dict[str, str]:
        headers = {
            "Accept": "application/vnd.github.v3+json",
        }
        if self.token:
            headers["Authorization"] = f"token {self.token}"
        if content_type:
            headers["Content-Type"] = content_type

        return headers

    def _request_internal(
        self,
        url: str,
        *,
        params: Optional[Dict[str, str]] = None,
        data: Optional[Dict[str, str]] = None,
        content: Optional[bytes] = None,
        request: Optional[Callable] = None,
        content_type: Optional[str] = None,
    ) -> httpx.Response:
        request = request or httpx.get
        headers = self._request_headers(content_type=content_type)
        kwargs = {}
        if data is not None:
            kwargs["json"] = data
        if content is not None:
            kwargs["content"] = content
        return request(
            url,
            headers=headers,
            params=params,
            follow_redirects=True,
            timeout=self.timeout,
            **kwargs,
        )

    def _request(
        self,
        api: str,
        *,
        params: Optional[Dict[str, str]] = None,
        data: Optional[Dict[str, str]] = None,
        request: Optional[Callable] = None,
    ) -> httpx.Response:
        return self._request_internal(
            f"{self.url}{api}", params=params, data=data, request=request
        )

    def _request_all(
        self,
        api: str,
        *,
        params: Optional[Dict[str, str]] = None,
        data: Optional[Dict[str, str]] = None,
        request: Optional[Callable] = None,
    ) -> Iterator[JSON]:
        response: httpx.Response = self._request(
            api, params=params, data=data, request=request
        )

        yield from response.json()

        next_url = _get_next_url(response)

        while next_url:
            response = self._request_internal(
                next_url, params=params, data=data, request=request
            )

            yield from response.json()

            next_url = _get_next_url(response)

        return data

    def _get_paged_items(
        self, api: str, key: str, *, params: Optional[JSON_OBJECT] = None
    ) -> Iterable[JSON_OBJECT]:
        """
        Internal method to get the paged items information from different REST
        URLs.
        """
        page = 1
        per_page = 100
        params = params or {}
        params.update({"per_page": per_page, "page": page})

        response: httpx.Response = self._request(
            api, request=httpx.get, params=params
        )
        response.raise_for_status()

        json = response.json()
        total = json.get("total_count", 0)
        items = json[key]
        downloaded = len(items)

        while total - downloaded > 0:
            page += 1
            params = {"per_page": per_page, "page": page}

            response = self._request(api, request=httpx.get, params=params)
            response.raise_for_status()

            json = response.json()
            items.extend(json[key])
            downloaded = len(items)

        return items



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=protected-access


import contextlib
import io
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos.version.cmake import CMakeVersionCommand, CMakeVersionParser
from pontos.version.helper import VersionError


class CMakeVersionCommandTestCase(unittest.TestCase):
    def test_raise_exception_file_not_exists(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = "CMakeLists.txt"
        fake_path.exists.return_value = False
        with self.assertRaises(VersionError):
            CMakeVersionCommand(project_file_path=fake_path)

    def test_raise_exception_no_project(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = "CMakeLists.txt"
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = ""
        with self.assertRaises(ValueError):
            CMakeVersionCommand(project_file_path=fake_path).run(args=["show"])
        fake_path.read_text.assert_called_with(encoding="utf-8")

    def test_raise_exception_file_not_found(self):
        with self.assertRaises(
            VersionError, msg="CMakeLists.txt file not found"
        ):
            CMakeVersionCommand()

    def test_return_error_string_incorrect_version_on_verify(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = "CMakeLists.txt"
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = (
            "project(VERSION so_much_version_so_much_wow)\n"
            "set(PROJECT_DEV_VERSION 0)"
        )

        result = CMakeVersionCommand(project_file_path=fake_path).run(
            args=["verify", "so_much_version_so_much_wow"]
        )
        self.assertTrue(
            isinstance(result, str), "expected result to be an error string"
        )
        self.assertEqual(
            result,
            "The version so_much_version_so_much_wow is not PEP 440 compliant.",
        )

    @patch(
        "pontos.version.cmake.CMakeVersionCommand.get_current_version",
        MagicMock(return_value="21.4"),
    )
    @patch("pontos.version.cmake.CMakeVersionCommand._print")
    def test_return_0_correct_version_on_verify(self, print_mock):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = "CMakeLists.txt"
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = ""

        result = CMakeVersionCommand(project_file_path=fake_path).run(
            args=["verify", "21.4"]
        )
        self.assertEqual(0, result)
        print_mock.assert_called_with("OK")

    def test_should_call_print_current_version_without_raising_exception(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = "CMakeLists.txt"
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = "project(VERSION 21)"
        CMakeVersionCommand(project_file_path=fake_path).run(args=["show"])
        fake_path.read_text.assert_called_with(encoding="utf-8")

    def test_raise_update_version(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = "CMakeLists.txt"
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = (
            "project(VERSION 21)\nset(PROJECT_DEV_VERSION 0)"
        )
        with contextlib.redirect_stdout(io.StringIO()):
            CMakeVersionCommand(project_file_path=fake_path).run(
                args=["update", "22", "--develop"]
            )
        fake_path.read_text.assert_called_with(encoding="utf-8")
        fake_path.write_text.assert_called_with(
            "project(VERSION 22)\nset(PROJECT_DEV_VERSION 1)", encoding="utf-8"
        )

    def test_update_version_equal_not_force(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = "CMakeLists.txt"
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = (
            "project(VERSION 22)\nset(PROJECT_DEV_VERSION 0)"
        )
        with contextlib.redirect_stdout(io.StringIO()) as buf:
            CMakeVersionCommand(project_file_path=fake_path).run(
                args=["update", "22", "--develop"]
            )
            self.assertEqual(buf.getvalue(), "Version is already up-to-date.\n")
            fake_path.read_text.assert_called_with(encoding="utf-8")


class CMakeVersionParserTestCase(unittest.TestCase):
    def test_get_current_version_single_line_project(self):
        under_test = CMakeVersionParser("project(VERSION 2.3.4)")
        self.assertEqual(under_test.get_current_version(), "2.3.4")

    def test_update_version_project(self):
        under_test = CMakeVersionParser("project(VERSION 2.3.4)")
        self.assertEqual(
            under_test.update_version("2.3.5"), "project(VERSION 2.3.5)"
        )

    def test_update_raise_exception_when_version_is_incorrect(self):
        under_test = CMakeVersionParser("project(VERSION 2.3.4)")
        with self.assertRaises(VersionError):
            under_test.update_version("so_much_version_so_much_wow")

    def test_not_confuse_version_outside_project(self):
        under_test = CMakeVersionParser(
            "non_project(VERSION 2.3.5)\nproject(VERSION 2.3.4)"
        )
        self.assertEqual(under_test.get_current_version(), "2.3.4")

    def test_get_current_version_multiline_project(self):
        under_test = CMakeVersionParser("project\n(\nVERSION\n\t    2.3.4)")
        self.assertEqual(under_test.get_current_version(), "2.3.4")

    def test_find_project_dev_version(self):
        test_cmake_lists = """
        project(
            DESCRIPTION something
            VERSION 41.41.41
            LANGUAGES c
        )
        set(
            PROJECT_DEV_VERSION 1
        )
        """
        under_test = CMakeVersionParser(test_cmake_lists)
        self.assertEqual(under_test._project_dev_version_line_number, 7)
        self.assertEqual(under_test._project_dev_version, "1")

    def test_update_project_dev_version(self):
        test_cmake_lists = """
        project(
            DESCRIPTION something
            VERSION 41.41.41
            LANGUAGES c
        )
        set(
            PROJECT_DEV_VERSION 1
        )
        """
        under_test = CMakeVersionParser(test_cmake_lists)

        self.assertEqual(under_test._project_dev_version_line_number, 7)
        self.assertEqual(under_test._project_dev_version, "1")
        result = under_test.update_version("41.41.41", develop=False)
        self.assertEqual(under_test._project_dev_version, "0")
        self.assertEqual(
            result,
            test_cmake_lists.replace(
                "PROJECT_DEV_VERSION 1", "PROJECT_DEV_VERSION 0"
            ),
        )

    def test_update_project_dev_version_when_succeeded_by_another_set(self):
        test_cmake_lists = """
        cmake_minimum_required(VERSION 3.1)

        project(hello_world VERSION 41.41.41)
        set(PROJECT_DEV_VERSION 1)

        add_executable(app main.c)
        """
        under_test = CMakeVersionParser(test_cmake_lists)

        self.assertEqual(under_test._project_dev_version_line_number, 4)
        self.assertEqual(under_test._project_dev_version, "1")
        result = under_test.update_version("41.41.41", develop=False)
        self.assertEqual(under_test._project_dev_version, "0")
        self.assertEqual(
            result,
            test_cmake_lists.replace(
                "PROJECT_DEV_VERSION 1", "PROJECT_DEV_VERSION 0"
            ),
        )

    def test_get_current_version_multiline_project_combined_token(self):
        under_test = CMakeVersionParser(
            "project\n(\nDESCRIPTION something VERSION 2.3.4 LANGUAGES c\n)"
        )
        self.assertEqual(under_test.get_current_version(), "2.3.4")

    def test_raise_exception_project_no_version(self):
        with self.assertRaises(ValueError) as context:
            CMakeVersionParser("project(DESCRIPTION something LANGUAGES c)")
        self.assertEqual(
            str(context.exception), "unable to find cmake version in project."
        )

    def test_raise_exception_no_project(self):
        with self.assertRaises(ValueError) as context:
            CMakeVersionParser(
                "non_project(VERSION 2.3.5)",
            )

        self.assertEqual(
            str(context.exception), "unable to find cmake version."
        )



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from argparse import Namespace
from pathlib import Path

import httpx

from pontos.github.api import GitHubRESTApi
from pontos.helper import shell_cmd_runner
from pontos.terminal import Terminal

from .helper import get_current_version, get_project_name


def sign(
    terminal: Terminal,
    args: Namespace,
    *,
    token: str,
    **_kwargs,
) -> bool:
    if not token and not args.dry_run:
        # dry run doesn't upload assets. therefore a token MAY NOT be required
        # for public repositories.
        terminal.error(
            "Token is missing. The GitHub token is required to upload "
            "signature files."
        )
        return False

    project: str = (
        args.project
        if args.project is not None
        else get_project_name(shell_cmd_runner)
    )
    space: str = args.space
    git_tag_prefix: str = args.git_tag_prefix
    release_version: str = (
        args.release_version
        if args.release_version is not None
        else get_current_version(terminal)
    )
    signing_key: str = args.signing_key

    git_version: str = f"{git_tag_prefix}{release_version}"
    repo = f"{space}/{project}"

    github = GitHubRESTApi(token=token)

    if not github.release_exists(repo, git_version):
        terminal.error(f"Release version {git_version} does not exist.")
        return False

    zip_destination = Path(f"{project}-{release_version}.zip")
    with github.download_release_zip(
        repo, git_version, zip_destination
    ) as zip_progress:
        terminal.download_progress(zip_progress)

    tarball_destination = Path(f"{project}-{release_version}.tar.gz")
    with github.download_release_tarball(
        repo, git_version, tarball_destination
    ) as tar_progress:
        terminal.download_progress(tar_progress)

    file_paths = [zip_destination, tarball_destination]

    for progress in github.download_release_assets(repo, git_version):
        file_paths.append(progress.destination)
        terminal.download_progress(progress)

    for file_path in file_paths:
        terminal.info(f"Signing {file_path}")

        if args.passphrase:
            process = shell_cmd_runner(
                f"gpg --pinentry-mode loopback --default-key {signing_key}"
                f" --yes --detach-sign --passphrase {args.passphrase}"
                f" --armor {file_path}"
            )
        else:
            process = shell_cmd_runner(
                f"gpg --default-key {signing_key} --yes --detach-sign "
                f"--armor {file_path}"
            )

        process.check_returncode()

    if args.dry_run:
        return True

    upload_files = [
        (Path(f"{str(p)}.asc"), "application/pgp-signature") for p in file_paths
    ]
    terminal.info(f"Uploading assets: {[str(p[0]) for p in upload_files]}")

    try:
        for uploaded_file in github.upload_release_assets(
            repo, git_version, upload_files
        ):
            terminal.ok(f"Uploaded: {uploaded_file}")
    except httpx.HTTPError as e:
        terminal.error(f"Failed uploading asset {e}")
        return False

    return True
